#include <stdio.h>
#include <stdlib.h>
#include "navigation.h"
#include "chart.h"
#include "colours.h"
#include "planet.h"
#include "clipping.h"
#include "commander.h"
#include "menu.h"
#include "eliteloop.h"

// add an offset to take into account the fact that each jump 
// requires a bit of effort to reach the station or sun scoop
#define HYPER_SPACE_FACTOR  (800)

// the current count of open and closed nodes 
int g_open_cnt;
int g_closed_cnt;

typedef struct {
  unsigned char index;  // planet index
  unsigned char parent; // parent planet (used to trace the final path)
  unsigned long g;      // real cost to get here
  unsigned long h;      // estimated cost to the end
} node_t;

#define SECTION_FLAGS  __attribute__ ((section (".sbss")))

node_t  g_open[256]  SECTION_FLAGS;
node_t  g_closed[256]  SECTION_FLAGS;
node_t  g_successors[128]  SECTION_FLAGS;

seed_t  g_nearest_planets[256] SECTION_FLAGS;
int g_near_cnt = 0;

// create the 256 planets
void calculate_planets() {
  seed_t cseed = g_commander.gal_seed;
  int i;
  g_near_cnt = 0;
  for (i = 0; i < 256; i++) {
    g_nearest_planets[g_near_cnt++] = cseed;
    churn_seed(&cseed);
    churn_seed(&cseed);
    churn_seed(&cseed);
    churn_seed(&cseed);
  }
}


int g_goal;
// estimate the distance from planet n to the goal
int heuristic(int n) 
{
  int dx = abs(g_nearest_planets[n].s3 - g_nearest_planets[g_goal].s3);
  int dy = abs(g_nearest_planets[n].s1 - g_nearest_planets[g_goal].s1)/2;

  return (dx*dx + dy*dy);
}

#if 0
#define OPEN(n) {g_open[(n)]; g_open_cnt++;}
#define IS_OPEN(n) (g_open[(n)]!=0)
#define CLOSE(n) {g_closed[(n)]; g_closed_cnt++;}
#define IS_CLOSED(n) (g_closed[(n)]!=0)
#endif
void open_node(const node_t * rhs) {
  g_open[g_open_cnt] = *rhs;
  g_open_cnt++;
}
void close_node(const node_t * rhs) {
  g_closed[g_closed_cnt] = *rhs;
  g_closed_cnt++;
}


int default_g(int index)
{
  return HYPER_SPACE_FACTOR;
}

int government_g(int index)
{
  static const int factor_gov [] = {
    1700, 950, 400, 300, 250, 150, 100, 0
  };
  // calculate the government
  int gov = (g_nearest_planets[index].s2/8)&7;
   
  return HYPER_SPACE_FACTOR + factor_gov[gov];
}
MenuFunctionPtr g_func = government_g;

// find all possible planets that can be jumped to from parent
int generate_succesors(const node_t * parent, int max_dist)
{
  int x = g_nearest_planets[parent->index].s3;
  int y = g_nearest_planets[parent->index].s1;
  int count = 0;
  int i;
  for (i = 0; i < 256; i++) {
    if (i != parent->index)
    {
      
      int dx = g_nearest_planets[i].s3 - x;
      int dy = (g_nearest_planets[i].s1 - y)/2;
      int dist = (dx*dx + dy*dy);
      if (dist <= max_dist)
      {
        // add to succesors, set cost
        g_successors[count].index = i;
        g_successors[count].parent = parent->index;
        g_successors[count++].g = parent->g + dist + g_func(i);
      }
    }
  }
  return count;
}

int is_open(int planet_id) {
  int i;
  for (i = 0; i < g_open_cnt; i++) {
    if (g_open[i].index == planet_id)
      return i;
  }
  return -1;
}
int is_closed(int planet_id) {
  int i;
  for (i = 0; i < g_closed_cnt; i++) {
    if (g_closed[i].index == planet_id)
      return i;
  }
  return -1;
}

void remove_from(node_t * table, int *count, int id)
{
  node_t * dest = &table[id];
  int i;
  *count = (*count ) - 1;
  for (i = id; i < *count; i++)
  {
    *dest = *(dest+1);
    dest++;
  }
}

void remove_node(int openid, int closedid)
{
  // remove planet id from open and closed lists
  if (openid != -1)
  {
    remove_from(g_open, &g_open_cnt, openid);
  }

  if (closedid != -1)
  {
    remove_from(g_closed, &g_closed_cnt, closedid);
  }
}

// returns a copy of the best current solution that we have yet to explore
const node_t get_lowest_f_open()
{
  int lowest_f = 0x7fffffff;
  int lowest_index = 0;
  int i; 
  for (i = 0; i < g_open_cnt; i++)
  {
    if ((g_open[i].g + g_open[i].h) <= lowest_f)
    {
      lowest_f = (g_open[i].g + g_open[i].h);
      lowest_index = i;
    }
  }
  
  return g_open[lowest_index];
}

#include "gbahandler.h"
// plot the path from the end to the start...
int draw_path(int to, int * jumps)
{
  int i;
  int destination_index = is_closed(to);
  if (destination_index == -1)
  {
    // planet cannot be reached in the current ship
    return -1;
  }
  unsigned int total = 0;
  for (i = 0; i < 256;i++)
  {
    int si = g_closed[destination_index].index;
    if (g_closed[destination_index].g == 0)
      break; //reached start
    int parent_index = is_closed(g_closed[destination_index].parent);
    if (parent_index == -1)
      return -1;
    int ei = g_closed[parent_index].index;

    coord_t start, end;
    if (g_my_stats.screen == SCR_SHORT_CHART) {
      start.x = (g_nearest_planets[si].s3 - g_commander.cpl_coord.x)*4 + SHORT_CENTRE_X;
      start.y = (g_nearest_planets[si].s1- g_commander.cpl_coord.y)*2 + SHORT_CENTRE_Y;
      end.x = (g_nearest_planets[ei].s3- g_commander.cpl_coord.x)*4 + SHORT_CENTRE_X;
      end.y = (g_nearest_planets[ei].s1- g_commander.cpl_coord.y)*2 + SHORT_CENTRE_Y;
      lineColour = col_blue;
    } else {
      start.x = g_nearest_planets[si].s3;
      start.y = g_nearest_planets[si].s1/2 + GAL_OFFSET_Y;
      
      end.x = g_nearest_planets[ei].s3;
      end.y = g_nearest_planets[ei].s1/2 + GAL_OFFSET_Y;
      if (g_func == government_g)
        lineColour = col_yellow;
      else
        lineColour = col_red;
    }
    int dx = (g_nearest_planets[si].s3 - g_nearest_planets[ei].s3);
    int dy = (g_nearest_planets[si].s1 - g_nearest_planets[ei].s1)/2;
    int dist = (dx*dx) + (dy*dy);
    total += bios_sqrt(dist)*4;

    
    if (g_my_stats.screen == SCR_SHORT_CHART) {
      draw_clipped_line(&start, &end);
    }else {
      draw_line(start.x, start.y, end.x, end.y);
    }
    destination_index = parent_index;
  }
  *jumps = i;
  return total;
}
  
// we want to go from planet "from" to planet "to"
// we have "fuel" for each hop
int path_find(galcoord_t fromc, galcoord_t toc, int fuel)
{
  /*if (g_func == government_g)
    g_func = default_g;
  else*/

  g_func = government_g;
    
  calculate_planets();
  int i;
  int from = -1;
  int to = -1;
  for (i = 0; i < 256; i++) {
    if (fromc.x == g_nearest_planets[i].s3 && fromc.y == g_nearest_planets[i].s1)
      from = i;
    if (toc.x == g_nearest_planets[i].s3 && toc.y == g_nearest_planets[i].s1)
      to = i;
  }
  if (from == -1 || to == -1)
  {
    return -1;
  }

  g_goal = to;
  g_open_cnt=0;
  g_closed_cnt=0;

  // place start planet on the open list
  g_successors[0].index = from; 
  g_successors[0].parent = from; 
  g_successors[0].g = 0; 
  g_successors[0].h = heuristic(from); 
  open_node(&g_successors[0]);

  // max distance accounts for truncation since we don't sqrt each distance
  int max_dist = (fuel*fuel)/16;
  max_dist = bios_sqrt(max_dist);
  max_dist++;
  max_dist = (max_dist*max_dist)-1;
  
  while (g_open_cnt) {
    const node_t node_current = get_lowest_f_open();

    // Add node_current to the CLOSED list
    close_node(&node_current);
    remove_node(is_open(node_current.index), -1);
    // if node_current is the same state as node_goal we have found the solution
    if (node_current.index == to) {
      break;
    }
    
    // Generate each state that can come after node_current
    int successors = generate_succesors(&node_current, max_dist);
    for (i = 0; i < successors; i++) 
    {
      // Set the cost of node_successor to be the cost of node_current 
      // plus the cost to get to node_successor from node_current
      //
      // this is already done in succesor generation
      
      // find node_successor on the OPEN list
      int openid = is_open(g_successors[i].index);

      //  if node_successor is on the OPEN list but the existing one is as 
      //  good or better then discard this successor and continue
      if (openid != -1 && 
          ( g_open[openid].g <= g_successors[i].g )
         )
      {
        continue;
      }

      int closedid = is_closed(g_successors[i].index);
      //  if node_successor is on the CLOSED list but the existing one is as 
      //  good or better then discard this successor and continue
      if (closedid != -1 && 
            ( g_closed[closedid].g <= g_successors[i].g )
         )
      {
        continue;
      }
      // else node_successor was not on the open or closed lists
      // or it was on the lists but this version is better

      // Remove occurences of node_successor from OPEN and CLOSED
      remove_node(openid, closedid);
      
      // Set the parent of node_successor to node_current
      // This is already done in generate_succesors

      // Set h to be the estimated distance to node_goal (Using the heuristic function)
      g_successors[i].h = heuristic(g_successors[i].index);

      // Add node_successor to the OPEN list
      open_node(&g_successors[i]);
    }
    // Sanity check - make sure we don't run out of nodes
    if (g_open_cnt == 256 || g_closed_cnt == 256)
      return -1;
  }

  return to;
}

