#include "options.h"
#include "interrupts.h"
#include "soundfx.h"


typedef struct 
{
  unsigned short length;
  unsigned char count;
  unsigned char update;
} channel_t;

// globals to let us know which of the 4 channels are used
// copy length in here, so that when == 0, can skip
static channel_t g_channel[4] = {};
int g_have_sound = 0;
// pointers to the data in each channel
// the data cannot be changed, but the pointer can
const soundentry_t *  g_sound_data[4] = {};

int maskTable[] = {
  0x1100,
  0x2200,
  0x4400,
  0x8800,
};

// for the 3 bits..
int sound3voltbl[] = {
  //000 - Mute
  //001 - 100%
  //100 - 75%
  //010 - 50%
  //011 - 25%
  0, // 0
  3, // 1  25
  2, // 2  50
  2, // 3  50
  4, // 4  75
  4, // 5  75
  1, // 6  100
  1, // 7  100
};

// interrupt called when sound timer fires
// set next sound effect (or turn off sound if none left)
void sound_ir() {
  // loop over the 4 channels and update the sound register 
  // if all 4 are off, then disable the sound interrupt (a timer)
  if (g_have_sound == 0) 
    return;
  int i;
  int have_sound = 0;
  for (i = 0; i < 4; i++) {
    // get the current value of the sound, mask out the 
    u16 sound =  REG_SGCNT0_L;
    u16 mask = maskTable[i];
    if (g_channel[i].length) {
      have_sound = 1;
      sound |= mask|0x77;
      REG_SGCNT0_L = sound;
      // do we need to change the frequency?
      if (g_channel[i].update) {
        // play the new sound 
        switch (i) {
          case 0:
            REG_SG10_H = ((g_sound_data[i]->volume)<<12)| SOUND1_SG10_H;
            REG_SG11 = g_sound_data[i]->freq;
            break;
          case 1:
            REG_SG20 = (((g_sound_data[i]->volume)<<12)|SOUND2_SG20);
            REG_SG21 = (g_sound_data[i]->freq);
            break;
          case 2:
            REG_SG30_H = (sound3voltbl[g_sound_data[i]->volume/2]<<12)|SOUND3_SG30_H;
            REG_SG31 = g_sound_data[i]->freq;
            break;
          case 3:
            REG_SG40 = (g_sound_data[i]->volume<<12)|SOUND4_SG40;
            REG_SG41 = g_sound_data[i]->freq;
            break;
        }
      }

      if (g_channel[i].count) {
        g_channel[i].count--;
      }
      if (!g_channel[i].count) {
        // This sound has ended. Cue the next one.
        g_channel[i].length--;
        // if length is 0, then the rest of this won't be checked next time anyway
        if (g_channel[i].length) {
          g_channel[i].count = g_sound_data[i]->count;
          g_channel[i].update = 1;
          g_sound_data[i]++;
        }
      }
    } else {
      sound &= ~mask;
      REG_SGCNT0_L = sound;
    }
  }

  if (!have_sound) {
    // turn off irq
    g_have_sound = 0;
    //REG_SGCNT1 = 0;
  }
}


// start the sound effect
// 
void play_sound(int id) 
{

  // make sure we don't interrupt the sound part way through or
  // it may write a random sound effect to a register
  SetInterrupt(0);

  int channel = g_elite_soundfx[id].channel;
  // if channel 0-2 and its preferred channel is in use 
  // then try another channel. oh! no, because lasers sounds screw up?

  g_channel[channel].length = g_elite_soundfx[id].length;
  g_sound_data[channel] = g_elite_soundfx[id].sound_data;
  // each sample can last a few ticks if need be
  g_channel[channel].count = g_sound_data[channel]->count;
  g_channel[channel].update = 1;

  // turn on the sound, if it is off
  g_have_sound = 1;
  REG_SGCNT1 = SOUND_MASTER_ENABLE;
  SetInterrupt(1);
}

// initialise the gbc sound registers
void init_sound() 
{
  REG_SGCNT1 = SOUND_MASTER_ENABLE; // only allowed to write this bit  // 0x4000084
  REG_SGCNT0_H |= SOUND_OUTPUT_FULL; // 0x4000082
  REG_SGCNT0_L |= 0x77;  // full sound, left and right // 0x4000080

  // sound 1
  REG_SG10_L = 0x8;  // sweep off
  // length is 8 bits on this one
  REG_SG10_H = SOUND1_SG10_H;

  // sound 2
  REG_SG20 = SOUND2_SG20;

  // sound 3
  REG_SG30_L = SOUND3BANK64 | SOUND3SETBANK0;
  REG_SG30_L |= SOUND3PLAY;
  REG_SG30_H =  SOUND3_SG30_H;
  // use a square wave
#if 1
  REG_SGWR0=0x77887788;
  REG_SGWR1=0x77887788;
  REG_SGWR2=0x77887788;
  REG_SGWR3=0x77887788;
#else
  REG_SGWR0=0x77777777;
  REG_SGWR1=0x88888888;
  REG_SGWR2=0x77777777;
  REG_SGWR3=0x88888888;
#endif
  REG_SG40 = SOUND4_SG40;

}

void play_song(int id) 
{
  sxmp_start_song(id, 1);
  if (g_game_options.music_off)
    sxmp_pause(1);
}
