
#include "input.h"
#include "bios_div.h"
#include <stdlib.h>
#include <string.h>

#ifdef USE_MALLOC
#include <malloc.h>
#endif

#include "universe.h"
#include "shipdata.h"
#include "planet.h"
#include "navigation.h"
#include "gbahandler.h"
#include "scanner.h"
#include "launch.h"
#include "tactics.h"

#include "intro2.h"
#include "eliteloop.h"
#include "gfx.h"
#include "soup.h"
#include "flying.h"
#include "commander.h"
#include "market.h"
#include "upgrade.h"
#include "soundfx.h"
#include "missions.h"
#include "stars.h"

#define RIGHT_ON_COMMANDER   0x65

// this value was 128 in "Classic Elite". Altered here to avoid some of the
// oddness when getting very close to the space station
#define MINIMUM_DOCK_DISTANCE   170

#ifdef USE_MALLOC
stellarobject_t * g_objects = 0;
int * g_object_types = 0;
#else
stellarobject_t g_objects[NUM_SHIPS] __attribute__ ((section (".sbss")));
int g_object_types[MAX_SHIP_ID] __attribute__ ((section (".sbss")));
#endif
//int g_dangerous_ships = 0;

const matrix_t default_rotmat = {
  {ROTMAT_1,0,       0        },
  {0       ,ROTMAT_1,0        },
  {0       ,0       ,-ROTMAT_1},
};

void init_universe(void) {
#ifdef USE_MALLOC
  g_objects = (stellarobject_t *) malloc(NUM_SHIPS * sizeof(stellarobject_t));
#endif
  
#ifdef USE_MALLOC
  g_object_types = (int *)malloc(MAX_SHIP_ID * sizeof(int));
#endif
  zero_objects();
  g_need_combat_seq = 1;
  tmp_save_commander = default_commander;
  g_my_stats = default_stats;
}


// code at $370d, or line 4500 Elite E
int add_new_ship(int shipid) {
  int i;
  if (shipid > 0) {
    // the ship may not be present in this collection...
    // TODO: alter the ship collections to have more variety without gaps
    if (!ships[shipid-1]) {
      return -1;
    }
  }
  int new_index = -1;
  for (i = 0; i < NUM_SHIPS; i++) {
    if (g_objects[i].type == 0) {
      new_index = i;
      break;
    }
  }
  if (new_index == -1) {
    return -1;
  }

  g_objects[new_index].speed = 0;
  g_objects[new_index].accel = 0;
  if (shipid > 0) {
    // non planet
    g_objects[new_index].energy = ships[shipid-1]->max_energy;

    // load missile count
    g_objects[new_index].missiles = ships[shipid-1]->arsenal & 7;
    
    // sort out the flags
    g_objects[new_index].flag =
      current_ship_flags[shipid-1]&~(REMOVE_OBJECT|FLAG_DOCKING);

    g_objects[new_index].ai_state = 0;
    
    g_object_types[shipid] += 1;
    // elite A only..
    if (shipid >= SHIP_ESCAPE_CAPSULE && shipid < SHIP_BOA)
      g_my_stats.space_junk++;
    
  }

  g_objects[new_index].type = (char)shipid;
  g_objects[new_index].target = 0;
  // set the rotation matrix
  for (i = 0; i < 3; i++) {
    g_objects[new_index].rotmat[i] = default_rotmat[i];
  }
  return new_index;
}

// add a new ship and copy the contents of parent to it
int copy_to_new_ship(int type, stellarobject_t * parent) {
  int child = add_new_ship(type);
  if (child != -1 && parent != 0) {
    //memcpy(&g_objects[child],parent, sizeof(stellarobject_t));
    g_objects[child].rotmat[0] = parent->rotmat[0];
    g_objects[child].rotmat[1] = parent->rotmat[1];
    g_objects[child].rotmat[2] = parent->rotmat[2];
    g_objects[child].location = parent->location;
    g_objects[child].rotx = parent->rotx;
    g_objects[child].rotz = parent->rotz;
    g_objects[child].speed = parent->speed;
    g_objects[child].missiles |= parent->missiles;
    g_objects[child].flag |= parent->flag;
    g_objects[child].ai_state |= parent->ai_state;
    
    g_objects[child].target = 0;
  }
  return child;
}

void zero_objects(void) {
  int i,j;
  for (i = 0; i < NUM_SHIPS; i++) {
    memset(&g_objects[i],0, sizeof(stellarobject_t));
    for (j = 0; j < 3; j++)
      g_objects[i].rotmat[j] = default_rotmat[j];
  }
  clear_ship_counts();
  g_my_stats.space_junk = 0;
  // no space station, so remove it
  deactivate_sprites(SST_SPRITE, SST_SPRITE+1);
}

void docking_injury(stellarobject_t * station) {
  // 1420 etc
  if (g_delta > 5) {
    oops(0x80,station);
    return;
  } 
  // scrape the sides?... 1426
  int dmg = (player_beta&4)^5;
  oops(dmg,station);

}

int has_docked(stellarobject_t * station) {

  if (station->flag & FLAG_ANGRY) {
    //didn't dock, SST angry
    return 0;
  }

  // auto_pilot shouldn't blow us up!
  if (g_my_stats.auto_pilot&1)
    return 1;

  if (station->rotmat[2].z > -0x5600) {  // 56 -> 0.33
    return 0;   
  }

  // get stations position as normalised vector 4283  
  int x = station->location.x;
  int y = station->location.y;
  int z = station->location.z;
  int total = abs(x)|abs(y)|abs(z);
  while (total < 0x10000) {
    total*=2;
    x *= 2;
    y *= 2;
    z *= 2;
  }
  vector32_t v = { (x/256), (y/256), (z/256)};
  norm(&v);

  if (v.z < 0x56 ) {
    return 0;   
  }

  if (abs(station->rotmat[1].x) < 0x5000) {
    return 0;   
  }

  return 1;
  
}

void remove_ship(int i) {
  // remove ship from universe 3D80
  int type = g_objects[i].type;
  // check missile lock - if we have one and it's this ship
  // need to print target lost
  if ( g_my_stats.missile_armed > 0 && 
       g_my_stats.missile_armed == i 
      ) 
  {
    flight_message(0xc8);
  }

  g_objects[i].type = 0;
  g_objects[i].location.x = 
    g_objects[i].location.y = 
    g_objects[i].location.z = 0;
  g_objects[i].rotx =
    g_objects[i].rotz = 0;

  // was i our missile target?
  int k;
  for (k = 0; k < NUM_SHIPS; k++)
  {
    if (k != i) {
      stellarobject_t * obj = &g_objects[k];
      if ( (obj->type == SHIP_MISSILE) && 
          !(obj->ai_state & ATTACK_PLAYER) && 
           (obj->ai_state&0x3f) == i // target == removed ship
           ) 
      {
        // our missile was after this ship, so die...
        obj->missiles |= REMOVE_OBJECT;
      }
    }
  }
  
  if (type == SHIP_SST) {
    // 3D31 remove space station - clear S, add sun etc
    // clear big S
    deactivate_sprites(SST_SPRITE, SST_SPRITE+1);
    g_object_types[SHIP_SST] = 0;
    add_new_ship(SUN);
    g_objects[1].location.y = 0x060000;
  } else {
    // 3D99
    if (type == SHIP_UNKOWN) {
      // mission completed if funny ship
      g_commander.mission |= MISSION_1_DONE;
    }
    if (type >= SHIP_ESCAPE_CAPSULE && 
        type <= SHIP_SPLINTER ) {
      g_my_stats.space_junk--;
    } else {
      if (type != SHIP_MISSILE) {
        g_my_stats.ships_present--;
        if (g_my_stats.ships_present<0)
          g_my_stats.ships_present=0;
      }
    }
    g_object_types[type]--;
  }

}

#ifdef HAVE_C64_KILLS
// these are consistent with the values on http://www.iancgbell.clara.net/elite/faq.htm#A4
// each shipheader_t has its own index to this table
/*
static const unsigned short points_table[] ={
  0x0000, 
  0x0095, // SHIP_MISSILE
  0x0000, // SHIP_SST
  0x0010, // SHIP_ESCAPE_CAPSULE
  0x000a, // SHIP_ALLOY
  0x000a, // SHIP_CARGO_CANNISTER
  0x0006, // SHIP_BOULDER
  0x0008, // SHIP_ASTEROID
  0x000a, // SHIP_SPLINTER
  0x0010, // SHIP_SHUTTLE
  0x0011, // SHIP_TRANSPORTER
  0x00ea, // SHIP_COBRA3 (traders)
  0x00aa, // PYTHON
  0x00d5, // BOA
  0x0100, // SHIP_ANACONDA
  0x0055, // SHIP_WORM <- SIDEWINDER
  0x001a, // SHIP_VIPER
  0x0055, // SHIP_COBRA_MK1 (weak pirates)
  0x0080, // MAMBA
  0x0055, // KRAIT
  0x005a, // ADDER
  0x0055, // GECKO
  0x00aa, // SHIP_PYTHON (strong pirates COBRA_I too)
  0x0032, // WORM
  0x012a, // COBRA 3   (bounty hunters/"lone wolves")
  0x0115, // ASP_II
  0x012a, // PYTHON
  0x0140, // SHIP_FER_DE_LANCE 
  0x00c0, // MORAY
  0x02aa, // SHIP_THARGOID
  0x0021, // SHIP_THARG/LET
  0x0555, // SHIP_UNKOWN
};
*/
#endif

void increment_kills(int type) {
  
  unsigned int kills = g_commander.kills;
#ifdef HAVE_C64_KILLS
  //int index = ships[type-1]->kill_value;
  g_commander.kills += ships[type-1]->kill_value;//points_table[index];
  if  ( (kills>>16) < (g_commander.kills>>16) ) 
#else
  g_commander.kills++;
  if  ( (kills>>8) < (g_commander.kills>>8) ) 
#endif
  {
    flight_message(RIGHT_ON_COMMANDER);
  }
}

int damage_ship(int ship, int damage) {
  // 0x80 mask used for 'not player doing the damage'

  int is_player = !(ship&0x80);
  ship &= 0x7f;
  
  // $14F4
  int effect = (ships[g_objects[ship].type-1]->max_energy&7)
    - damage;
  if (effect < 0) {
    int en  = g_objects[ship].energy + effect;
    if (en <= 0) {
      g_objects[ship].energy = 0;
      if (!(g_objects[ship].missiles&EXPLODE_OBJECT)) {
        // In space nobody hears you scream. But this is only a video game, so they do. 
        play_sound(SOUNDFX_EXPLOSION);
        g_objects[ship].missiles |= REMOVE_OBJECT;
        if (is_player) // doing the damage
          increment_kills(g_objects[ship].type);
        else
          g_objects[ship].target = 0x80;
      } 
      return 1;
    } else {
      g_objects[ship].energy = en;
    }
  }
  return 0;
}

int hit_check(stellarobject_t * obj) {
  vector32_t * objloc = &obj->location;

  if (objloc->z < 0 || obj->type < 0) {
    return 0;
  }
  if ( (obj->missiles&EXPLODE_OBJECT) 
      || abs(objloc->y)>0xff 
      || abs(objloc->x)>0xff ) {
    return 0;
  }

  int x2 = objloc->x*objloc->x;
  int y2 = objloc->y*objloc->y;
  int size2 = x2+y2;
  if (size2 > ships[obj->type-1]->target_size) {
    return 0;
  }
  return 1;
}

void collide_with(int i) {
  // 142F
  // damage other ship for 0x40...
  // this is in stark contrast to TNK, where collisions
  // blew up the enemy craft
  damage_ship(i,0x40);
  make_angry(i);
  // take a whack - this is actually safer than TNK
  // in that you lost 128 + object.energy/2
  oops(128,&g_objects[i]);
}

void scoop_item(int i, int cargo) {
  //   if no space it explodes, causing damage
  //   else set 80 in flag of item to remove it
  //2AA3
  int has_space = 1;
  if (cargo == ALIEN_ITEMS || cargo < TONNE_ITEMS) {
    int total = g_commander.cargo[ALIEN_ITEMS];
    int i;
    for (i = 0; i < TONNE_ITEMS; i++) {
      total += g_commander.cargo[i];
    }
    if (total >= g_commander.cargo_space)
      has_space = 0;
  } else {
    // kg, g etc
    if (g_commander.cargo[cargo] >= 0xff) {
      has_space = 0;
    }
  }

  if (has_space) {
    // $13E6
    play_sound(SOUNDFX_SCOOP);
    g_commander.cargo[cargo]++;
    flight_message(FOOD_STRING+cargo);
    g_objects[i].flag |= REMOVE_OBJECT;
  }
  else {
    collide_with(i);
  }
}

void check_scoops(int i) {
  if (g_objects[i].location.y < 0 && g_commander.equipment[FUEL_SCOOPS]) {
    // scoop branch (fitted and thing is Y<0) 
    // is cargo canister?
    if (g_objects[i].type == SHIP_CARGO_CANNISTER) {
      //   scoop item - cargo is random
      scoop_item(i,get_rand()&0xf);
    } else {
      // else is scoopable? 
      int cargo = ships[g_objects[i].type-1]->cargo/16;
      if (cargo == 0) {
        // if not make the ship angry
        collide_with(i);
      } else {
        scoop_item(i,cargo+1);
      }
    }
  } else {
    // else not scoops
    // cause damage, make angry
    collide_with(i);
  }
}

void check_remove(int i) {
  // is this correct? It uses the hideous BVS and I'm in a muddle about it
  // checks if the ai flag has top bit set (AI control) or not (is stupid)
  // if stupid, don't do anything...
  //if (g_objects[i].ai_state & 0x80)  // where did this come from?
  if (g_objects[i].flag & FLAG_INNOCENT ) { // 0x40 | 0x20
    // pay the consequences for that...
    int fine = g_objects[i].flag & FLAG_DEFENCELESS;
    if (!fine) {
      // he was asking for it...
      fine = 1;
    } else {
      // oh man, you blew up an escape pod!
      // add 0x40 to the legal status, or fill in the bit closest
      // to bit 0x40..
      while (fine&g_commander.legal_status)
        fine/=2;
    }

    // g_commander.legal_status |= fine; // correct?
    // seems to add if not carry set, else don't change anything
    if ( (g_commander.legal_status + fine) < 256)
      g_commander.legal_status += fine;
  } else {
    // bounty - Elite A shafts you here...
    // it refuses to give bounty when a message is shown. 
    // If "ENERGY LOW" is shown you don't get any bounty!
    // 0x80 for if someone else blew it up
    if (g_objects[i].target != 0x80 ) {
      const shipheader_t * const data = ships[g_objects[i].type-1];
      if ( (g_my_stats.miss_jump == 0) && data && (g_my_stats.combat_seq==0)) {
        int bounty = data->bountylo | (data->bountyhi<<8);
        if (bounty > 0) {
          g_commander.cash += bounty;
          flight_message(0); // Cash 0.9 Cr
        }
      }
    }
  }
  //
}

void switch_to_view(stellarobject_t * ship) {
  if (g_my_stats.view == FRONT_VIEW)
    return;
  else if (g_my_stats.view == REAR_VIEW) {
    // rear view is simple
    ship->location.x = -ship->location.x;
    ship->location.z = -ship->location.z;

    int i;
    for (i = 0; i < 3; i++) {
      ship->rotmat[i].x = -ship->rotmat[i].x;
      ship->rotmat[i].z = -ship->rotmat[i].z;
    }
  } else {
    int x_sign = g_my_stats.view==RIGHT_VIEW?1:-1;
    int z_sign = g_my_stats.view==RIGHT_VIEW?-1:1;

    int tmp = ship->location.x;
    ship->location.x = ship->location.z*z_sign;
    ship->location.z = tmp*x_sign;

    int i;
    for (i = 0; i < 3; i++) {
      int tmp = ship->rotmat[i].x;
      ship->rotmat[i].x = ship->rotmat[i].z*z_sign;
      ship->rotmat[i].z = tmp*x_sign;
    }

  }
}

// if rand_in is not set, then create new random amount
// possible early return if top bit of rand not set
void release_random(int type, int parent, int rand_in) {
  //$164C
  int r = rand_in;
  if (!rand_in) {
    r = get_rand();
    if (!(r&0x80))
      return;
    r = r&ships[g_objects[parent].type]->cargo&0xf;
  }
  while (r) {
    // was carrying cargo... launch type from parent with no AI.
    if (-1 == launch_ship_from_other(type, &g_objects[parent], 0) )
    {
      // failed to allocate, give up
      break;
    }
    r--;
  }
}

void release_goodies(int i) {
  //if we kill the thing, throw out goodies $1482
  //        goodies for asteroid == alloys
  if (g_objects[i].type == SHIP_ASTEROID) {
    // check laser, if mining launch random&3 splinter objects
    if (g_my_stats.laser_strength == 
  player_ship_data[g_commander.ship].laser_strength[SHIPDATA_LASER_MINING]) 
    {
      int r = get_rand()&3;
      if (r) {
  release_random(SHIP_SPLINTER, i, r);
      }
    }
  } 
  // launch a random amount of ship type SHIP_ALLOY
  release_random(SHIP_ALLOY, i, 0);
  // launch a random amount of ship type SHIP_CARGO_CANNISTER
  release_random(SHIP_CARGO_CANNISTER, i, 0);
}

int distancecmp(const void * a, const void * b)
{
  stellarobject_t * oa = &g_objects[*(int*)a];
  stellarobject_t * ob = &g_objects[*(int*)b];
  if( oa->distance < ob->distance)
    return -1;
  if( oa->distance > ob->distance)
    return 1;
  return 0;
}

int ycmp(const void * a, const void * b)
{
  stellarobject_t * oa = &g_objects[*(int*)a];
  stellarobject_t * ob = &g_objects[*(int*)b];
  if( oa->location.y < ob->location.y)
    return -1;
  if( oa->location.y > ob->location.y)
    return 1;
  return 0;
}

void update_universe() {
  reset_lollipops();
  int i;
  int order_index = 0;
  int order[NUM_SHIPS];
  // first draw the static star field background
  int need_dust = 1; // do we still need to draw the dust field
  if (g_my_stats.screen == SCR_LAUNCHED){
    update_stars(0);
  }
  int hit_anything = 0;
  for (i = 0; i < NUM_SHIPS; i++) {
    if (g_objects[i].type != 0) {
      if (g_objects[i].type > 0 ) {
        const shipheader_t * const data = ships[g_objects[i].type-1];
        if (data) {
          mveit(&g_objects[i], data, i);

          if ( (g_objects[i].ai_state&0x80) 
              && !(g_objects[i].missiles&(EXPLODE_OBJECT|REMOVE_OBJECT))
              && ((i^g_mcount)&7)==0 )  
          {
            // time for tactics 
            tactics(&g_objects[i], data);
          }
        }
        /*
        // also update scanner
        if (g_my_stats.screen == SCR_LAUNCHED  && !g_my_stats.game_over
              && !(g_objects[i].missiles&(EXPLODE_OBJECT|REMOVE_OBJECT))
            ) 
        { 
          scanner(i);
        }
        */
      }
      else {
        // planets sun
        mveit(&g_objects[i], 0, i);
      }

      //if (g_my_stats.game_over == 0 ) {
      // check distance of object

      g_objects[i].distance = abs(g_objects[i].location.x/256)|
        abs(g_objects[i].location.y/256)|
        abs(g_objects[i].location.z/256);
      int dist = 
        (g_objects[i].missiles&(EXPLODE_OBJECT|REMOVE_OBJECT))|
        g_objects[i].distance;
        /*abs(g_objects[i].location.x/256)|
        abs(g_objects[i].location.y/256)|
        abs(g_objects[i].location.z/256);*/
      if (dist == 0) {
        // very close
        dist = (abs(g_objects[i].location.x)&0xff)|
          (abs(g_objects[i].location.y)&0xff)|
          (abs(g_objects[i].location.z)&0xff);
        if (dist < MINIMUM_DOCK_DISTANCE) {
          if (g_objects[i].type == SHIP_SST) {
            if (has_docked(&g_objects[i])) {
              dock_player();
              break;
            } else {
              docking_injury(&g_objects[i]);
            }
          } else if (dist < 0x40 && g_objects[i].type != SHIP_MISSILE) {
            // check for scoopage, collisions
            // 13BA closer than 0x40  (dist&0xC0 == 0)
            // not missile
            check_scoops(i); 
            // TODO - sound effect for scoop
          }
        }
      }
      // MA26
      // 144A
      // check targetting for missiles
      // if not on view screen do not check
      // check view, flip object to view
      // Do HIT CHECK

      if (g_my_stats.screen == SCR_LAUNCHED) { 
        // flip to view
        stellarobject_t view_ship = g_objects[i];
        switch_to_view(&view_ship);

        int hit = hit_check(&view_ship);
        if (hit && g_my_stats.laser_on) {
          // if hit:
          //   if laser_on:
          //     noise (crunch?)
          hit_anything = view_ship.location.z;
          play_sound(SOUNDFX_CRUNCH);
          //     if type = 2, make station angry and end
          make_angry(i);
          if (g_objects[i].type != SHIP_SST) {
            //     othership.energy -= laser/2
            //     for thargoids, laser/4
            int damage = g_my_stats.laser_on/2;
            if (g_objects[i].type == SHIP_UNKOWN) {
              damage/=2;
            } 
            if (damage_ship(i,damage)) {
              // killed it!
              // this next line is to update the explosion flags:
              view_ship.missiles = g_objects[i].missiles;
              release_goodies(i);
            }
          }
        } 
        if (hit && g_my_stats.missile_armed == -1) {
          // lock target
          g_my_stats.missile_armed = i;
          //   beep and target miss
          lock_missile();
          if (g_commander.equipment[IFF_SYSTEM]) {
            // show ship name
            const shipheader_t * const data = ships[g_objects[i].type-1];
            if (data) {
              flight_message( IFF_FLIGHT_MESSAGE( data->name ) );
            }
          }
        }

        if (g_objects[i].type > 0 ) {
          //const shipheader_t * const data = ships[g_objects[i].type-1];
          order[order_index++] = i;
          /*
          entry(&view_ship, data);
          // update flags and expl count, just in case
          g_objects[i].missiles = view_ship.missiles;
          g_objects[i].expl_count = view_ship.expl_count;
          g_objects[i].r0         = view_ship.r0;
          g_objects[i].r1         = view_ship.r1;
          g_objects[i].r2         = view_ship.r2;
          g_objects[i].r3         = view_ship.r3;
          */
        } else if (g_objects[i].type < 0) {
          draw_planet(&view_ship);
        }
      }

      // MA15 - update energy, 
      // check for removal
      if (g_objects[i].type > 0) {
        if (g_objects[i].flag&REMOVE_OBJECT) {
          // removal the easy way..
          remove_ship(i);
          continue;
        }
        if (!(g_objects[i].missiles&REMOVE_OBJECT) || 
            !(g_objects[i].missiles&EXPLODE_OBJECT) ) {
          // check for large distance
          if ( abs(g_objects[i].location.x) > 0xE000 ||
              abs(g_objects[i].location.y) > 0xE000 ||
              abs(g_objects[i].location.z) > 0xE000) 
          {
            // out of range 14f1
            remove_ship(i);
            continue;
          }
        } else {
          // exploded and ready for removal
          check_remove(i);
          remove_ship(i);
          continue;
        }
      } // type > 0
      //} // game over
    } // type != 0
  }
  if (hit_anything!=0) {
    activate_laser_hit(hit_anything);
  } else {
    deactivate_sprites(LASER_HIT_SPRITE,LASER_HIT_SPRITE+1);
  }
  // draw in correct z-order
  if (order_index) {
    qsort(order, order_index, sizeof(int), distancecmp );
    for (i = order_index-1; i >=0; i--) {
      int j = order[i];
      if (g_objects[j].type > 0) {
        const shipheader_t * const data = ships[g_objects[j].type-1];
        stellarobject_t view_ship = g_objects[j];
        switch_to_view (&view_ship);
        if (need_dust && view_ship.distance < 6) {
          // draw the close-in particle effect
          update_stars(1);
          need_dust = 0;
        }
        entry(&view_ship, data);
        g_objects[j].missiles   = view_ship.missiles;
        g_objects[j].expl_count = view_ship.expl_count;
        g_objects[j].r0         = view_ship.r0;
        g_objects[j].r1         = view_ship.r1;
        g_objects[j].r2         = view_ship.r2;
        g_objects[j].r3         = view_ship.r3;
      }
    }

    /* 
     * Draw objects with y < 0 below the scanner, NES Elite style. Objects must
     * be sorted by Y order, those with y < 0 have priority below scanner
     * background but also must be 'below' those with y >= 0. Sprite to sprite
     * priority is determined only by the OAM index, which is why the objects
     * are ordered here.
     */
    qsort(order, order_index, sizeof(int), ycmp );
    for (i = 0; i < order_index; i++) {
      int j = order[i];
      if (g_my_stats.screen == SCR_LAUNCHED  && !g_my_stats.game_over
          && !(g_objects[j].missiles&(EXPLODE_OBJECT|REMOVE_OBJECT))
         ) 
      { 
        scanner(j);
      }
    }
  }
  // check to see if we still need the dust field (e.g. when no ships were present)
  if (g_my_stats.screen == SCR_LAUNCHED && need_dust) {
    update_stars(1);
  }

  
}

void add_planet(void) {
  int planet_type = (cpl_info.tech&2)?CRATER:BEACHBALL;
    
  int planetid =  add_new_ship(planet_type);
  g_objects[planetid].rotz =
    g_objects[planetid].rotx = 0x7f;

  g_objects[planetid].location.x = g_objects[planetid].location.y = 0;
  g_objects[planetid].location.z = 0x010000;
} 
void add_station(void) {
  g_objects[1].type = 0;
  int id = add_new_ship(SHIP_SST);
  g_objects[id].rotz = 0;
  g_objects[id].rotx = 0x7f;
  g_objects[id].ai_state = 0x81;

  g_objects[id].rotmat[2].x = -g_objects[id].rotmat[2].x;
  g_objects[id].rotmat[2].y = -g_objects[id].rotmat[2].y;
  g_objects[id].rotmat[2].z = -g_objects[id].rotmat[2].z;
  
  // make station angry if legal status is > 127
  if (g_commander.legal_status > 127)
    g_objects[id].flag |= FLAG_ANGRY;
  g_objects[id].location.x = g_objects[id].location.y = 0;
  g_objects[id].location.z = -0x0100;

  // add the big S
  if (g_my_stats.screen == SCR_LAUNCHED) 
    activate_sprites(SST_SPRITE, SST_SPRITE+1);

} 

// pointer to shipset key, or 0 if should generate the key
void do_load_shipset(int * set) 
{
  
  // load the random ship set for this planet
  // this takes care of variety of pirates, etc.
  // code from $4233, data converted from Elite-A.ssd
  int x;
  if (set == 0)
    x = g_commander.galaxy + 
    (cpl_info.gov>=3)*2 + (cpl_info.tech>=10) + (get_rand()&3)*4;
  else
    x = *set;
  g_my_stats.shipset_key = x;
#ifdef HAVE_GBA_SHIPSETS
  // ship sets are generated 'on the fly' for the gba version
  // only a subset of shiptypes is allowed in each position
  // e.g. only pirate-like ships can be pirates, only vipers can be vipers
  // but allows swapping of space station type without doing
  // "if ship == station type 1 or ship == station type 2 ...."
  // and "if shiptype is a cobra or a mamba or a....  "

  // load the shipset
  int i;
  int result, remainder;
  for (i = 0; i < 32; i++) {
    // load the default ship first - for 11/30 this is all we have to do
    ships[i] = ship_collections[i]->ships[0];
    if ( ship_collections[i]->length > 1) {
      // remainder = x%number of ships
      FastIntDivide(x, ship_collections[i]->length, &result,&remainder);
      switch (i+1) {
        case SHIP_SST:
          {
            ships[i] = ship_collections[i]->ships[x&1];
          }
          break;
        case SHIP_SHUTTLE:
          {
            if (x == 11) {
              ships[i] = ship_collections[i]->ships[1];
            }
          }
          break;
        default:
          {
            // the default behaviour is to just load a random ship
            ships[i] = ship_collections[i]->ships[remainder];
          }
          break;
      }
    } else {
      switch (i+1) {
        case SHIP_UNKOWN:
          // only allow the ship at the right planet
          if (g_commander.cpl_coord.x != 0x90 || 
              g_commander.cpl_coord.y != 0x21) {
            ships[i] = 0;
          } 
          break;
        default : break;
      }
    }
    // the flags have either 1 or 2 entries
    if (ship_flags[i][1] == 0) {
      current_ship_flags[i] = ship_flags[i][0];
    }
    else {
      // mostly this only effects if the ship has ecm or not
      current_ship_flags[i] = ship_flags[i][x&1];
    }
  }
  
#else
  // every 4 sets there is a pattern
  // gov*2 + tech>9
  // everything offset by galaxy
  ships = ship_collections[x];
  current_ship_flags = ship_flags[x];


/*  x = 1;
  ships = ship_collections[x];
  current_ship_flags = ship_flags[x];*/
#endif
}

// load a new shipset, we don't care which
void load_shipset(void)
{
  do_load_shipset(0);
}

/*
void  generate_random_location(int i) {
  // 3E60
  g_objects[i].rotmat[0] = default_rotmat[0];
  g_objects[i].rotmat[1] = default_rotmat[1];
  g_objects[i].rotmat[2] = default_rotmat[2];

  int r1, r2;
  get_2rand(&r1,&r2);
  int lyub = (r1/8);
  int lxub = (r2&0x1f);
  g_objects[i].location.x = r1|(lxub*256);
  g_objects[i].location.y = r2|(lyub*256);
  
  if (r1&1)
    g_objects[i].location.x = -g_objects[i].location.x;
  if (r1&2)
    g_objects[i].location.y = -g_objects[i].location.y;

  g_objects[i].location.z = (( 0x50 - lyub - lxub )&0xff)*256;
  
}
*/

void  generate_random_location(stellarobject_t * obj) {
  // 3E60
  obj->rotmat[0] = default_rotmat[0];
  obj->rotmat[1] = default_rotmat[1];
  obj->rotmat[2] = default_rotmat[2];

  int r1, r2;
  get_2rand(&r1,&r2);
  int lyub = (r1/8);
  int lxub = (r2&0x1f);
  obj->location.x = r1|(lxub*256);
  obj->location.y = r2|(lyub*256);
  
  if (r1&1)
    obj->location.x = -obj->location.x;
  if (r1&2)
    obj->location.y = -obj->location.y;

  obj->location.z = (( 0x50 - lyub - lxub )&0xff)*256;
  
}

void  clear_ship_counts(void) {
  int i;
  for (i = 0; i < MAX_SHIP_ID; i++) {
    g_object_types[i] = 0;
  }
  g_my_stats.ships_present=0;
}

